from typing import Any, cast

import torch

from elasticai.creator.base_modules.conv1d import Conv1d as Conv1dBase
from elasticai.creator.nn.fixed_point._math_operations import MathOperations
from elasticai.creator.nn.fixed_point._two_complement_fixed_point_config import (
    FixedPointConfig,
)
from elasticai.creator.nn.fixed_point.conv1d.design import Conv1d as Conv1dDesign
from elasticai.creator.vhdl.design_creator import DesignCreator


class BatchNormedConv1d(DesignCreator, torch.nn.Module):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        in_channels: int,
        out_channels: int,
        signal_length: int,
        kernel_size: int | tuple[int],
        bn_eps: float = 1e-5,
        bn_momentum: float = 0.1,
        bn_affine: bool = True,
        stride: int | tuple[int] = 1,
        padding: int | tuple[int] = 0,
        bias: bool = True,
        device: Any = None,
    ) -> None:
        super().__init__()
        self._config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)
        self._operations = MathOperations(config=self._config)
        self._signal_length = signal_length
        self._conv1d = Conv1dBase(
            operations=self._operations,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
            device=device,
        )
        self._batch_norm = torch.nn.BatchNorm1d(
            num_features=out_channels,
            eps=bn_eps,
            momentum=bn_momentum,
            affine=bn_affine,
            track_running_stats=True,
            device=device,
        )

    @property
    def conv_weight(self) -> torch.Tensor:
        return self._conv1d.weight

    @property
    def conv_bias(self) -> torch.Tensor | None:
        return self._conv1d.bias

    @property
    def bn_weight(self) -> torch.Tensor:
        return self._batch_norm.weight

    @property
    def bn_bias(self) -> torch.Tensor:
        return self._batch_norm.bias

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        has_batches = x.dim() == 2
        input_shape = (
            (x.shape[0], self._conv1d.in_channels, -1)
            if has_batches
            else (1, self._conv1d.in_channels, -1)
        )
        output_shape = (x.shape[0], -1) if has_batches else (-1,)

        x = x.view(*input_shape)
        x = self._conv1d(x)
        x = self._batch_norm(x)
        x = self._operations.quantize(x)

        return x.view(*output_shape)

    def create_design(self, name: str) -> Conv1dDesign:
        def float_to_signed_int(value: float | list) -> int | list:
            if isinstance(value, list):
                return list(map(float_to_signed_int, value))
            return self._operations.config.as_integer(value)

        def flatten_tuple(x: int | tuple[int, ...]) -> int:
            return x[0] if isinstance(x, tuple) else x

        bn_mean = cast(torch.Tensor, self._batch_norm.running_mean)
        bn_variance = cast(torch.Tensor, self._batch_norm.running_var)
        bn_epsilon = self._batch_norm.eps
        conv_weight = self._conv1d.weight
        conv_bias = (
            torch.tensor([0] * self._conv1d.out_channels)
            if self._conv1d.bias is None
            else self._conv1d.bias
        )

        std = torch.sqrt(bn_variance + bn_epsilon)
        weights = conv_weight / std
        bias = (conv_bias - bn_mean) / std

        if self._batch_norm.affine:
            weights = (self._batch_norm.weight * weights.t()).t()
            bias = self._batch_norm.weight * bias + self._batch_norm.bias

        return Conv1dDesign(
            name=name,
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            in_channels=self._conv1d.in_channels,
            out_channels=self._conv1d.out_channels,
            signal_length=self._signal_length,
            kernel_size=flatten_tuple(self._conv1d.kernel_size),
            weights=cast(list[list[list[int]]], float_to_signed_int(weights.tolist())),
            bias=cast(list[int], float_to_signed_int(bias.tolist())),
        )



from collections.abc import Iterable, Iterator, Mapping
from dataclasses import dataclass
from itertools import repeat
from string import Template as StringTemplate
from typing import Protocol, cast

from .resource_utils import read_text


def module_to_package(module: str) -> str:
    return ".".join(module.split(".")[:-1])


class Template(Protocol):
    parameters: dict[str, str | list[str]]
    content: list[str]


@dataclass
class InProjectTemplate(Template):
    package: str
    file_name: str
    parameters: dict[str, str | list[str]]

    def __post_init__(self) -> None:
        self.content = list(read_text(self.package, self.file_name))


class TemplateExpander:
    def __init__(self, template: Template) -> None:
        super().__init__()
        self._template = template

    def lines(self) -> list[str]:
        self._assert_all_variables_to_fill_exists()

        single_line_params, multi_line_params = _split_single_and_multiline_parameters(
            self._template.parameters
        )
        lines = _expand_template(self._template.content, **single_line_params)
        lines = _expand_multiline_template(lines, **multi_line_params)
        return list(lines)

    def unfilled_variables(self) -> set[str]:
        template_variables = _extract_template_variables(self._template.content)
        variables_to_fill = set(self._template.parameters.keys())
        return template_variables - variables_to_fill

    def _assert_all_variables_to_fill_exists(self) -> None:
        template_variables = _extract_template_variables(self._template.content)
        variables_to_fill = set(self._template.parameters.keys())
        not_existing_variables = variables_to_fill - template_variables
        if len(not_existing_variables) > 0:
            raise KeyError(
                "The template has no variables named"
                f" {', '.join(not_existing_variables)} to fill."
            )


def _split_single_and_multiline_parameters(
    parameters: Mapping[str, str | Iterable[str]],
) -> tuple[dict[str, str], dict[str, Iterator[str]]]:
    single_line_parameters = dict(
        cast(
            Iterator[tuple[str, str]],
            filter(lambda i: isinstance(i[1], str), parameters.items()),
        )
    )
    multiline_parameters = dict(
        cast(
            Iterator[tuple[str, Iterator[str]]],
            filter(lambda i: not isinstance(i[1], str), parameters.items()),
        )
    )
    return single_line_parameters, multiline_parameters


def _expand_multiline_template(
    template: str | Iterable[str], **kwargs: Iterable[str]
) -> Iterator[str]:
    """Expand a template field to multiple lines, while keeping indentation.
    Example:
        >>> template = "\\t$my_key"
        >>> values = ["hello,", "world", "!"]
        >>> "\\n".join(_expand_multiline_template(template, my_key=values))
        '\\thello,\\n\\tworld\\n\\t!'
    """
    lines = _unify_template_datatype(template)
    for line in lines:
        contains_no_key = True
        for key in kwargs:
            if f"${key}" in line:
                contains_no_key = False
                for placeholder_line, value in zip(repeat(line), kwargs[key]):
                    t = StringTemplate(placeholder_line)
                    yield t.safe_substitute({key: value})
                break
        if contains_no_key:
            yield line


def _unify_template_datatype(template: str | Iterable[str]) -> Iterator[str]:
    if isinstance(template, str):
        lines = template.splitlines()
    else:
        lines = list(template)
    yield from lines


def _expand_template(template: str | Iterable[str], **kwargs: str) -> Iterator[str]:
    for line in _unify_template_datatype(template):
        yield StringTemplate(line).safe_substitute(kwargs)


def _extract_template_variables(template: list[str]) -> set[str]:
    return set(StringTemplate("\n".join(template)).get_identifiers())



def parse_report(text: str):
    lines = text.split("\n")[:-2]

    def split_first_five_colons(line):
        all_split = line.split(":")
        reassembled = all_split[0:5] + [":".join(all_split[5:])]
        return reassembled

    separated = [split_first_five_colons(line) for line in lines]
    fields = ("source", "line", "column", "time", "type", "content")

    def parse_line(line):
        parsed = dict(zip(fields, line))
        parsed["type"] = parsed["type"][1:-1]
        parsed["time"] = parsed["time"][1:]
        parsed["line"] = int(parsed["line"])
        parsed["column"] = int(parsed["column"])
        return parsed

    return [parse_line(line) for line in separated]



from collections.abc import Iterator
from importlib import resources
from pathlib import Path, PurePath
from typing import ContextManager

PathType = str | PurePath
Package = resources.Package


def get_file_from_package(package: Package, file_name: str) -> ContextManager[Path]:
    """
    This is a context manager, because the returned file might be extracted from a zip file and the context manager
    will take care of removing the resulting temporary files on __exit__
    """
    for resource in resources.files(package).iterdir():
        if resource.name == file_name:
            return resources.as_file(resource)
    raise FileNotFoundError(
        f"The file '{file_name}' in package '{package}' does not exist."
    )


def read_text(package: Package, file_name: str) -> Iterator[str]:
    with get_file_from_package(package, file_name) as file:
        with open(file, "r") as opened_file:
            yield from map(lambda line: line.rstrip("\n"), opened_file)


def _read_bytes(package: Package, file_name: str) -> Iterator[bytes]:
    with get_file_from_package(package, file_name) as file:
        with open(file, "rb") as opened_file:
            yield from opened_file


def copy_file(package: Package, file_name: str, destination: PathType) -> None:
    with open(destination, "wb") as out_file:
        for data in _read_bytes(package=package, file_name=file_name):
            out_file.write(data)


def get_full_path(package: Package, file_name: str) -> str:
    with get_file_from_package(package, file_name) as file:
        return file.as_posix()


def read_text_from_path(path: PathType, encoding: str = "utf-8") -> str:
    with open(path, "r", encoding=encoding) as inp_file:
        return inp_file.read()


def save_text_to_path(text: str, path: PathType, encoding: str = "utf-8") -> None:
    with open(path, "w", encoding=encoding) as out_file:
        out_file.write(text)



from typing import Any, Protocol

from torch import Tensor
from torch.nn import Conv1d as _Conv1d
from torch.nn.functional import conv1d

from elasticai.creator.base_modules.math_operations import Quantize


class MathOperations(Quantize, Protocol): ...


class Conv1d(_Conv1d):
    def __init__(
        self,
        operations: MathOperations,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int],
        stride: int | tuple[int] = 1,
        padding: int | tuple[int] | str = 0,
        dilation: int | tuple[int] = 1,
        groups: int = 1,
        bias: bool = True,
        device: Any = None,
        dtype: Any = None,
    ) -> None:
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode="zeros",
            device=device,
            dtype=dtype,
        )
        self._operations = operations

    def forward(self, x: Tensor) -> Tensor:
        quantized_weights = self._operations.quantize(self.weight)
        quantized_bias = (
            self._operations.quantize(self.bias) if self.bias is not None else None
        )
        convolved = conv1d(
            input=x,
            weight=quantized_weights,
            bias=quantized_bias,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            groups=self.groups,
        )
        return self._operations.quantize(convolved)



import dataclasses

from ._number_conversion import (
    bits_to_integer,
    bits_to_rational,
    convert_rational_to_bit_pattern,
    integer_to_bits,
    max_integer,
    max_natural,
    max_rational,
    min_integer,
    min_natural,
    min_rational,
)


@dataclasses.dataclass
class FXPParams:
    total_bits: int
    frac_bits: int


class NumberConverter:
    def __init__(self, fxp_params: FXPParams):
        self._fxp_params = fxp_params

    def bits_to_integer(self, pattern: str) -> int:
        return bits_to_integer(pattern)

    def bits_to_rational(self, pattern: str) -> float:
        return bits_to_rational(pattern, frac_bits=self._fxp_params.frac_bits)

    def rational_to_bits(self, rational: float) -> str:
        return convert_rational_to_bit_pattern(
            rational=rational,
            total_bits=self._fxp_params.total_bits,
            frac_bits=self._fxp_params.frac_bits,
        )

    def bits_to_natural(self, pattern: str) -> int:
        return bits_to_rational(pattern, frac_bits=self._fxp_params.frac_bits)

    def integer_to_bits(self, number: int) -> str:
        return integer_to_bits(number, total_bits=self._fxp_params.total_bits)

    @property
    def max_rational(self) -> float:
        return max_rational(
            total_bits=self._fxp_params.total_bits, frac_bits=self._fxp_params.frac_bits
        )

    @property
    def max_integer(self) -> int:
        return max_integer(total_bits=self._fxp_params.total_bits)

    @property
    def min_rational(self) -> float:
        return min_rational(
            total_bits=self._fxp_params.total_bits, frac_bits=self._fxp_params.frac_bits
        )

    @property
    def min_integer(self) -> int:
        return min_integer(total_bits=self._fxp_params.total_bits)

    @property
    def max_natural(self) -> int:
        return max_natural(total_bits=self._fxp_params.total_bits)

    @property
    def min_natural(self) -> int:
        return min_natural(total_bits=self._fxp_params.total_bits)



from typing import cast

import torch

from elasticai.creator.base_modules.conv1d import MathOperations as Conv1dOps
from elasticai.creator.base_modules.linear import MathOperations as LinearOps
from elasticai.creator.base_modules.lstm_cell import MathOperations as LSTMOps

from ._round_to_float import RoundToFloat


class MathOperations(LinearOps, Conv1dOps, LSTMOps):
    def __init__(self, mantissa_bits: int, exponent_bits: int) -> None:
        self.mantissa_bits = mantissa_bits
        self.exponent_bits = exponent_bits

    @property
    def largest_positive_value(self) -> float:
        exponent_bias = 2 ** (self.exponent_bits - 1)
        return (2 - 1 / 2**self.mantissa_bits) * 2 ** (
            2**self.exponent_bits - exponent_bias - 1
        )

    @property
    def smallest_negative_value(self) -> float:
        return -self.largest_positive_value

    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return self._round(self._clamp(a))

    def _clamp(self, a: torch.Tensor) -> torch.Tensor:
        return torch.clamp(
            a, min=self.smallest_negative_value, max=self.largest_positive_value
        )

    def _round(self, a: torch.Tensor) -> torch.Tensor:
        return cast(
            torch.Tensor,
            RoundToFloat.apply(a, self.mantissa_bits, self.exponent_bits),
        )

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a + b)

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.matmul(a, b))

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a * b)



from torch import Tensor

from ._math_operations import MathOperations as _FlpOperations


def quantize(x: Tensor, mantissa_bits: int, exponent_bits: int) -> Tensor:
    return _FlpOperations(
        mantissa_bits=mantissa_bits, exponent_bits=exponent_bits
    ).quantize(x)



import glob
import os
import subprocess

from ._ghdl_report_parsing import parse_report


class GHDLSimulator:
    """Run a simulation tool for a given `top_design` and save whatever is written to stdout
    for subsequent inspection.

    This runner uses the GHDL tool.
    The parsed content has the following keys: `("source", "line", "column", "time", "type", "content")'
    """

    def __init__(self, workdir, top_design_name):
        self._root = workdir
        self._ghdl_dir = "ghdl_build"
        self._files = list(glob.glob(f"**/*.vhd", root_dir=self._root, recursive=True))
        self._standard = "08"
        self._result = {}
        self._test_bench_name = top_design_name

    def initialize(self):
        """Call this function once before calling `run()` and on every file change."""
        os.makedirs(f"{self._root}/{self._ghdl_dir}", exist_ok=True)
        self._load_files()
        self._compile()

    def run(self):
        """Runs the simulation and saves whatever the tool wrote to stdout.
        You're supposed to call `initialize` once, before calling `run`."""
        self._result = self._execute_command_and_return_stdout(
            self._assemble_command("-r") + [self._test_bench_name]
        )

    def getReportedContent(self) -> list[str]:
        """Strips any information that the simulation tool added automatically to the output
        to return only the information that was printed to stdout via VHDL/Verilog statements.
        """
        parsed = parse_report(self._result)
        return list(line["content"] for line in parsed)

    def getFullReport(self) -> list[dict]:
        """Parses the output from the simulation tool, to provide a more structured representation.
        The exact content depends on the simulation tool.
        """
        return parse_report(self._result)

    def getRawResult(self) -> str:
        """Returns the raw stdout output as written by the simulation tool."""
        return self._result

    def _load_files(self):
        self._execute_command(self._assemble_command("-i") + self._files)

    def _compile(self):
        self._execute_command(
            self._assemble_command("-m") + ["-fsynopsys", self._test_bench_name]
        )

    def _execute_command(self, command):
        return subprocess.run(command, cwd=self._root)

    def _execute_command_and_return_stdout(self, command):
        return subprocess.run(
            command, cwd=self._root, capture_output=True
        ).stdout.decode()

    def _assemble_command(self, command_flag):
        return ["ghdl", command_flag, f"--std={self._standard}", self._workdir_flag]

    @property
    def _workdir_flag(self):
        return f"--workdir={self._ghdl_dir}"



from dataclasses import dataclass
from typing import Protocol, TypeVar, Union, cast, overload, runtime_checkable

T = TypeVar("T", bound="ConvertableToFixedPointValues")


@runtime_checkable
class ConvertableToFixedPointValues(Protocol[T]):
    def round(self: T) -> T: ...

    def int(self: T) -> T: ...

    def float(self: T) -> T: ...

    def __gt__(self: T, other: Union[int, float, T]) -> T:  # type: ignore
        ...

    def __lt__(self: T, other: Union[int, float, T]) -> T:  # type: ignore
        ...

    def __or__(self: T, other: T) -> T: ...

    def __mul__(self: T, other: Union[int, T, float]) -> T:  # type: ignore
        ...

    def __truediv__(self: T, other: Union[int, float]) -> T:  # type: ignore
        ...


@dataclass
class FixedPointConfig:
    total_bits: int
    frac_bits: int

    @property
    def minimum_as_integer(self) -> int:
        return 2 ** (self.total_bits - 1) * (-1)

    @property
    def maximum_as_integer(self) -> int:
        return 2 ** (self.total_bits - 1) - 1

    @property
    def minimum_as_rational(self) -> float:
        return -1 * (1 << (self.total_bits - 1)) / (1 << self.frac_bits)

    @property
    def maximum_as_rational(self) -> float:
        return int("1" * (self.total_bits - 1), 2) / (1 << self.frac_bits)

    def integer_out_of_bounds(self, number: T) -> T:
        return (number < self.minimum_as_integer) | (number > self.maximum_as_integer)

    def rational_out_of_bounds(self, number: T) -> T:
        return (number < self.minimum_as_rational) | (number > self.maximum_as_rational)

    @overload
    def as_integer(self, number: float | int) -> int: ...

    @overload
    def as_integer(self, number: T) -> T: ...

    def as_integer(self, number: float | int | T) -> int | T:
        if isinstance(number, ConvertableToFixedPointValues):
            return self._convert_T_to_integer(cast(T, number))
        return self._convert_float_or_int_to_integer(number)

    @overload
    def as_rational(self, number: float | int) -> float: ...

    @overload
    def as_rational(self, number: T) -> T: ...

    def as_rational(self, number: float | int | T) -> float | T:
        return number / (1 << self.frac_bits)

    def _convert_T_to_integer(self, number: T) -> T:
        return (number * (1 << self.frac_bits)).int().float()

    def _convert_float_or_int_to_integer(self, number: float | int) -> int:
        return round(number * (1 << self.frac_bits))



from elasticai.creator.vhdl.design.signal import Signal


def x(width: int) -> Signal:
    return Signal(name="x", width=width)


def done() -> Signal:
    return Signal(name="done", width=0)


def enable() -> Signal:
    return Signal(name="enable", width=0)


def clock() -> Signal:
    return Signal(name="clock", width=0)


def y(width: int) -> Signal:
    return Signal(name="y", width=width)


def x_address(width: int) -> Signal:
    return Signal(name="x_address", width=width)


def y_address(width: int) -> Signal:
    return Signal(name="y_address", width=width)



from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import InProjectTemplate


class MacTestBench:
    def __init__(self, uut, name, uut_name):
        self._uut = uut
        self._uut_name = uut_name
        self._inputs = None
        self._destination = None
        self.name = name

    def set_inputs(self, *inputs):
        self._inputs = inputs

    def parse_reported_content(self, outputs: list[str]):
        return 2 * int(outputs[0]) - 1

    @property
    def _width(self) -> str:
        return str(len(self._inputs[0]))

    def save_to(self, destination: Path):
        self._destination = destination
        inputs = self._prepare_inputs_for_test_bench(self._inputs)
        test_bench = InProjectTemplate(
            package="elasticai.creator.nn.binary.mac",
            file_name="testbench.tpl.vhd",
            parameters=inputs
            | {
                "uut_name": self._uut_name,
                "name": self.name,
                "total_width": self._width,
            },
        )
        self._uut.save_to(destination)
        destination.create_subpath(self.name).as_file(".vhd").write(test_bench)

    def _prepare_inputs_for_test_bench(self, inputs):
        x1, x2 = inputs

        def zero_one(xs):
            return ["0" if x < 0 else "1" for x in xs]

        def to_string(xs):
            return '"{}"'.format("".join(xs))

        x1 = zero_one(x1)
        x2 = zero_one(x2)
        inputs = {
            "x1": to_string(x1),
            "x2": to_string(x2),
        }
        return inputs



from elasticai.creator.file_generation.savable import Savable
from elasticai.creator.vhdl.simulated_layer import SimulatedLayer

from .._math_operations import MathOperations
from .design import MacDesign
from .mactestbench import MacTestBench


class MacLayer:
    def __init__(self, vector_width: int):
        self.ops = MathOperations()
        self.name = "bin_mac"
        self._vector_width = vector_width

    def __call__(self, a, b):
        return self.ops.matmul(a, b)

    def create_design(self, name: str) -> Savable:
        return MacDesign(vector_width=self._vector_width, name=name)

    def create_testbench(self, name: str) -> MacTestBench:
        return MacTestBench(
            uut=self.create_design("mac_wrapper_test"),
            uut_name="mac_wrapper_test",
            name=name,
        )

    def create_simulation(self, simulator, working_dir):
        return SimulatedLayer(self, simulator, working_dir=working_dir)



from torch import Tensor

from ._math_operations import MathOperations as _FxpOperations
from ._two_complement_fixed_point_config import FixedPointConfig as _FxpConfig


def quantize(x: Tensor, total_bits: int, frac_bits: int) -> Tensor:
    return _FxpOperations(
        config=_FxpConfig(total_bits=total_bits, frac_bits=frac_bits)
    ).quantize(x)
